/*  Test of the gtk editor widget.
 *  Copyright (C) 1998, 2000 Thomas Mailund <mailund@daimi.au.dk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include "gtkeditor.h"
#include <string.h>

#define BUF_MAX_SIZE 128	/* seems fair for one line */

/* global vars */
static GtkEditor *mainbuffer;
static GtkNotebook *book;
static GtkEntry *minibuffer;
static void (*minibuffer_callback)(char *) = NULL;
static GtkEditorScanner *g_cscanner = NULL;

/* minibuffer stuff */
static gint
minibuf_key_press (GtkWidget *widget, GdkEventKey *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->keyval == GDK_Return) {
    gtk_widget_grab_focus (GTK_WIDGET (mainbuffer));
    if (minibuffer_callback) {
      /* The callback is called with a copy of the text. It needs to
       * take care of freeing the text, should that be done. */
      minibuffer_callback (g_strdup(gtk_entry_get_text (minibuffer)));
    }
    gtk_entry_set_text (minibuffer, "");
  }

  return FALSE;
}

static void
get_text_from_minibuffer (GtkSignalFunc callback)
{
  gtk_entry_set_text (minibuffer, "");
  minibuffer_callback = callback;
  gtk_widget_grab_focus (GTK_WIDGET (minibuffer));
}

/* callbacks */
static void
load_file_cb (char *filename)
{
  int i;
  char buf[BUF_MAX_SIZE];
  FILE *fp = fopen (filename, "r");

#ifdef DEBUG
  g_print ("load_file_cb (%s)\n", filename);
#endif

  if (!fp) return;

  /* clear old text */
  i = gtk_text_get_length (GTK_TEXT(mainbuffer));
  if (i>0) {
    gtk_text_set_point (GTK_TEXT(mainbuffer), i);
    gtk_text_backward_delete (GTK_TEXT(mainbuffer), i);
  }

  /* load new */
  gtk_text_freeze (GTK_TEXT (mainbuffer));
  
  do {
    if (fgets (buf, BUF_MAX_SIZE, fp) == NULL) {
      break;
    } else {
      gtk_text_insert (GTK_TEXT (mainbuffer), NULL, NULL, NULL,
		       buf, strlen (buf));
    }
  } while (!feof (fp));

  gtk_text_thaw (GTK_TEXT (mainbuffer));

  /* clean up and hilite */
  fclose (fp);
  g_free (filename);
  /* FIXME: couldn't this be done while inserting? */
  gtk_editor_hilite_buffer (mainbuffer);
}

static void
save_file_cb (char *filename)
{
  int i, len;
  char *buf;
  FILE *fp = fopen (filename, "w");

#ifdef DEBUG
  g_print ("save_file_cb (%s)\n", filename);
#endif

  if (!fp) return;

  gtk_text_freeze (GTK_TEXT (mainbuffer));

  /* get_chars allocs a new buffer each time, so I have to do all this
   * g_free stuff */  
  len = gtk_text_get_length (GTK_TEXT (mainbuffer));
  for (i = 0; i < len - BUF_MAX_SIZE;) {
    buf = gtk_editable_get_chars (GTK_EDITABLE (mainbuffer), i, i + BUF_MAX_SIZE);
    fputs (buf, fp);
    i += BUF_MAX_SIZE;
    g_free (buf);
  }
  /* and the leftovers */
  buf = gtk_editable_get_chars (GTK_EDITABLE (mainbuffer), i, len);
  fputs (buf, fp);
  g_free (buf);

  gtk_text_thaw (GTK_TEXT (mainbuffer));

  fclose (fp);
}


static void
clear_buffer (void)
{
  gtk_editable_delete_text (GTK_EDITABLE (mainbuffer), 0,
			    gtk_text_get_length (GTK_TEXT (mainbuffer)));
}

static void
load_file (void)
{
  get_text_from_minibuffer (load_file_cb);
}

static void
save_file (void)
{
  get_text_from_minibuffer (save_file_cb);
}


/* hilite pattern test */
static void
install_pat (GtkEditor *cbuff)
{
  static GdkColor blue, greenish, blueish, red, brownish;
  static GList *pentries = NULL, *props, *bentries;
  static GdkColormap *cmap = NULL;
  static GdkFont *keywordsfont = NULL, *commentfont = NULL, *regularfont =NULL;
  static int cpp_comment_end[] = {11}, c_comment_end[] = {8}, string_end[] = {10, 11};

  if(g_cscanner) {    		
    gtk_editor_install_scanner (cbuff, g_cscanner, pentries);
    gtk_editor_install_bentries (cbuff, bentries);
    gtk_editor_install_text_properties (cbuff, props);
    return;
  }

  cmap = gtk_widget_get_colormap (GTK_WIDGET (mainbuffer));
  commentfont = gdk_font_load ("-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*");
  keywordsfont = gdk_font_load ("-adobe-courier-bold-r-normal-*-12-*-*-*-*-*-*-*");
  regularfont = gdk_font_load ("-adobe-courier-medium-r-normal-*-12-*-*-*-*-*-*-*");
  blue.pixel = 0;
  blue.red = 0;
  blue.green = 0;
  blue.blue = 65535;

  greenish.pixel = 0;
  greenish.red = 30000;
  greenish.green = 45000;
  greenish.blue = 0;

  blueish.pixel = 0;
  blueish.red = 40000;
  blueish.green = 0;
  blueish.blue = 55000;

  red.pixel = 0;
  red.red = 65535;
  red.green = 0;
  red.blue = 0;

  brownish.pixel = 0;
  brownish.red = 65535;
  brownish.green = 43908;
  brownish.blue = 16383;

  gdk_color_alloc (cmap, &blue);
  gdk_color_alloc (cmap, &greenish);
  gdk_color_alloc (cmap, &blueish);
  gdk_color_alloc (cmap, &red);
  gdk_color_alloc (cmap, &brownish);

  props = gtk_editor_text_property_new(regularfont, NULL, NULL,
          gtk_editor_text_property_new(commentfont, &red, NULL,
          gtk_editor_text_property_new(regularfont, &brownish, NULL,
          gtk_editor_text_property_new(keywordsfont, &blue, NULL,
          gtk_editor_text_property_new(regularfont, &greenish, NULL,
          gtk_editor_text_property_new(regularfont, &blueish, NULL, NULL))))));

  pentries = gtk_editor_pentry_new ("chars", "\"'\"(.*\"'\"|.*)", 0, 2,
	     gtk_editor_pentry_new ("keywords",
				    "(while|for|if|else|return|do|switch)/(([ \\t]*\\()|[^a-zA-Z0-9_]?)",
				    0, 3,
             gtk_editor_pentry_new ("types",
				    "int|long|float|char|void|static|const|extern|unsigned",
				    0, 4,
	     gtk_editor_pentry_new ("include",
				    "(#[ \\t]*include)[ \\t]*((\"\\\"\"[^<>\\n]*\"\\\"\"?)|(\"<\"[^\\\"\\n]*\">\"?))",
                                    0, 5,
	     gtk_editor_pentry_new ("cpp",
				    "[ \\t]*#[ \\t]*(include|if|ifdef|else|define|endif)",
				    0, 5,
	     gtk_editor_pentry_new ("C-functions", "([a-zA-Z_][a-zA-Z0-9_]*)/[ \\t]*\\(", 0, 5,
             gtk_editor_pentry_new ("C++-functions",
		                    "(([a-zA-Z_][a-zA-Z0-9_]*)::([~a-zA-Z_][a-zA-Z0-9_]*))[ \\t]*/\\(",
				    0, 5,
	     gtk_editor_pentry_new ("begin_c_comment", "\"/*\"", 0, 1,
	     gtk_editor_pentry_new ("end_c_comment", "\"*/\"", 0, 0,
	     gtk_editor_pentry_new ("begin_cpp_comment", "\"//\"", 0, 1,
	     gtk_editor_pentry_new ("quote", "\\\"", 0, 2,
	     gtk_editor_pentry_new ("new_line", "[\\n]", 0, 0,
	     gtk_editor_pentry_new ("divide", "\"/\"", 0, 0,
	     gtk_editor_pentry_new ("multiply", "\"*\"", 0, 0,
	     gtk_editor_pentry_new ("regular text",
				    "[ \\t]*|[a-zA-Z0-9_]*|[\\\\+=\\-~`!@#$\\%%^&()<>\\[\\]\\{\\}\\|:;?.,\\\']*",
				    0, 0, NULL)))))))))))))));
				
  /* test of blocks */
  bentries = gtk_editor_bentry_new ("C-comment", 7, 1, c_comment_end, 0, FALSE, 1,
             gtk_editor_bentry_new ("C++-comment", 9, 1, cpp_comment_end, 0, FALSE, 1,
             gtk_editor_bentry_new ("string", 10, 2, string_end, 0, FALSE, 2, NULL)));

  /* entry new copies font, so free it here */
  if (commentfont)
    gdk_font_unref (commentfont);

  /* entry_new copies font so we should free it */
  if (keywordsfont)
    gdk_font_unref (keywordsfont);

  g_cscanner = gtk_editor_compile_scanner ("c-scanner", pentries);
  gtk_editor_install_scanner (cbuff, g_cscanner, pentries);
  gtk_editor_install_bentries (cbuff, bentries);
  gtk_editor_install_text_properties (cbuff, props);
}

static void
new_cbuff (void)
{
  GtkWidget *hbox = gtk_hbox_new (FALSE, 0);
  GtkWidget *label = gtk_label_new ("C buffer");
  GtkWidget *editor = gtk_editor_new (NULL, NULL);
  GtkWidget *scrollbar = gtk_vscrollbar_new (GTK_TEXT(editor)->vadj);

  gtk_box_pack_start (GTK_BOX (hbox), editor, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), scrollbar, FALSE, FALSE, 0);

  gtk_text_set_editable (GTK_TEXT (editor), TRUE);

  install_pat (GTK_EDITOR (editor));

  gtk_widget_show (hbox);
  gtk_widget_show (label);
  gtk_widget_show (editor);
  gtk_widget_show (scrollbar);
  gtk_notebook_append_page (book, hbox, label);
}

static void
book_switch_page (GtkNotebook     *nbook,
		  GtkNotebookPage *page)
{
  GtkBoxChild *child = GTK_BOX (page->child)->children->data;

  mainbuffer = GTK_EDITOR (child->widget);
}


static void
quit (void)
{
  gtk_exit (0);
}

/* main program */
static void
display_greeting (GtkEditor *e)
{
  gtk_text_insert (GTK_TEXT (e), NULL, NULL, NULL,
		   "/* Welcome to this little editor test program.\n"
		   " *\n"
		   " * This text should be highlighted as a comment\n"
		   " * and the program below should like nice and coloured.\n"
		   " * If it does not, there's something wrong, and you should\n"
		   " * inform me on email <mailund@daimi.au.dk>\n"
		   " *\n"
		   " * Now, enjoy hacking with the editor widget.\n"
		   " *\n"
		   " *\t/mailund\n"
		   " */\n"
		   "\n"
		   "#include <stdio.h>\n"
		   "\n"
		   "int\n"
		   "main (int argc, char *argv[])\n"
		   "{\n"
		   "  if (argc < 1) {\n"
		   "    printf (\"hello, world!\\n\");\n"
		   "  } else {\n"
		   "    printf (\"howdy, %s\\n\", argv[1]);\n"
		   "  }\n\n"
		   "  return 0;\n"
		   "}\n", -1);
  gtk_editable_set_position (GTK_EDITABLE (e), 0);

  gtk_editor_hilite_buffer (e);
}

static GtkWidget*
create_window ()
{
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *edbox;
  GtkWidget *notebook;
  GtkWidget *editor;
  GtkWidget *label;
  GtkWidget *entry;
  GtkWidget *hbox;
  GtkWidget *button;
  GtkWidget *scrollbar;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Editor Test");

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  /* -- notebook for buffers -- */
  notebook = gtk_notebook_new ();
  book = GTK_NOTEBOOK (notebook);
  gtk_widget_set_usize (notebook, 450, 250);
  gtk_widget_show (notebook);
  gtk_box_pack_start (GTK_BOX (vbox), notebook, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (notebook), "switch_page",
		      GTK_SIGNAL_FUNC (book_switch_page), NULL);

  /* -- first buffer -- */
  edbox = gtk_hbox_new (FALSE, 0);
  editor = gtk_editor_new(NULL, NULL);
  gtk_box_pack_start (GTK_BOX (edbox), editor, TRUE, TRUE, 0);
  scrollbar = gtk_vscrollbar_new(GTK_TEXT(editor)->vadj);
  gtk_box_pack_start (GTK_BOX (edbox), scrollbar, FALSE, FALSE, 0);

  gtk_signal_connect_after (GTK_OBJECT (editor), "realize",
			    GTK_SIGNAL_FUNC (install_pat),
			    editor);
  gtk_signal_connect_after (GTK_OBJECT (editor), "realize",
			    GTK_SIGNAL_FUNC (display_greeting), NULL);

  gtk_text_set_editable (GTK_TEXT (editor), TRUE);
  gtk_widget_show (edbox);
  gtk_widget_show (editor);
  gtk_widget_show (scrollbar);
  label = gtk_label_new ("C buffer");
  gtk_widget_show (label);
  gtk_notebook_append_page (book, edbox, label);

  /* -- minibuffer -- */

  entry = gtk_entry_new ();
  gtk_widget_show (entry);
  gtk_box_pack_start (GTK_BOX (vbox), entry, FALSE, FALSE, 0);

  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC (quit),
		      NULL);

  minibuffer = GTK_ENTRY (entry);
  gtk_signal_connect (GTK_OBJECT (minibuffer), "key_press_event",
		      GTK_SIGNAL_FUNC (minibuf_key_press),
		      NULL);
  mainbuffer = GTK_EDITOR (editor);

  /* -- buttons -- */
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

  button = gtk_button_new_with_label ("New C buffer");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (new_cbuff), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("Clear");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (clear_buffer), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("Load file");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (load_file), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("Save file");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (save_file), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  return window;
}

int
main (int argc, char *argv[]) 
{
  GtkWidget *main_window;

  /* initialize */
  gtk_init (&argc, &argv);
  main_window = create_window ();

  /* show */
  gtk_widget_show (main_window);
  gtk_widget_grab_focus (GTK_WIDGET (mainbuffer));

  /* args? */
  if (argc > 1) {
    load_file_cb (g_strdup (argv[1])); /* we need a copy 'cause load
					* frees the string */
  }
  
  /* main loop */
  gtk_main ();

  return EXIT_SUCCESS;
}
